@@ -11,6 +11,7 @@
 #![allow(dead_code)]
 
 use std::{
+    mem::MaybeUninit,
     os::fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd},
     sync::LazyLock,
 };
@@ -18,13 +19,13 @@ use std::{
 use libseccomp::ScmpSyscall;
 use nix::{
     errno::Errno,
-    fcntl::{AtFlags, OpenHow},
-    sys::socket::{bind, connect, SockFlag, SockaddrLike},
+    fcntl::{AtFlags, OFlag, OpenHow},
+    sys::socket::{bind, connect, Shutdown, SockFlag, SockaddrLike},
     unistd::UnlinkatFlags,
     NixPath,
 };
 
-use crate::{compat::RenameFlags, fs::fillrandom, path::XPath};
+use crate::{compat::RenameFlags, path::XPath, rng::randint, uts::UtsName};
 
 /// A platform-sized secure cookie
 ///
@@ -42,170 +43,208 @@ pub(crate) type Cookie = u64;
 pub(crate) fn getcookie() -> Result<Cookie, Errno> {
     #[cfg(target_pointer_width = "32")]
     {
-        const N: usize = 4;
-        let mut buf = [0u8; N];
-        fillrandom(&mut buf)?;
-        Ok(Cookie::from_le_bytes(buf))
+        randint(0u32..=u32::MAX)
     }
 
     #[cfg(target_pointer_width = "64")]
     {
-        const N: usize = 8;
-        let mut buf = [0u8; N];
-        fillrandom(&mut buf)?;
-        Ok(Cookie::from_le_bytes(buf))
+        randint(0u64..=u64::MAX)
     }
 }
 
 // These cookies are confined by seccomp for use with openat2(2).
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static OPENAT2_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static OPENAT2_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: socket(2) $arg3
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static SOCKET_COOKIE_ARG3: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: socket(2) $arg4
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static SOCKET_COOKIE_ARG4: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: socket(2) $arg5
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static SOCKET_COOKIE_ARG5: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: accept4(2) $arg4
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static ACCEPT4_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: accept4(2) $arg5
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static ACCEPT4_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: bind(2) $arg3
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static BIND_COOKIE_ARG3: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: bind(2) $arg4
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static BIND_COOKIE_ARG4: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: bind(2) $arg5
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static BIND_COOKIE_ARG5: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: connect(2) $arg3
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static CONNECT_COOKIE_ARG3: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: connect(2) $arg4
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static CONNECT_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 /// System call argument cookie: connect(2) $arg5
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub static CONNECT_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG2: LazyLock<Cookie> =
+/// System call argument cookie: shutdown(2) $arg2
+#[expect(clippy::disallowed_methods)]
+pub static SHUTDOWN_COOKIE_ARG2: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG3: LazyLock<Cookie> =
+/// System call argument cookie: shutdown(2) $arg3
+#[expect(clippy::disallowed_methods)]
+pub static SHUTDOWN_COOKIE_ARG3: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG4: LazyLock<Cookie> =
+/// System call argument cookie: shutdown(2) $arg4
+#[expect(clippy::disallowed_methods)]
+pub static SHUTDOWN_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static RENAMEAT2_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG2: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG3: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG4: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE64_COOKIE_ARG3: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG2: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG3: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG4: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE64_COOKIE_ARG3: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static UNLINKAT_COOKIE_ARG3: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static UNLINKAT_COOKIE_ARG4: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static UNLINKAT_COOKIE_ARG5: LazyLock<Cookie> =
-    LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
-pub(crate) static LINKAT_COOKIE_ARG5: LazyLock<Cookie> =
+/// System call argument cookie: shutdown(2) $arg5
+#[expect(clippy::disallowed_methods)]
+pub static SHUTDOWN_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 
+#[expect(clippy::disallowed_methods)]
+pub(crate) static MEMFD_CREATE_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static MEMFD_CREATE_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static MEMFD_CREATE_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static MEMFD_CREATE_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static RENAMEAT2_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE64_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static TRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE64_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNAME_COOKIE_ARG1: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNAME_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNAME_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNAME_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNAME_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNLINKAT_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNLINKAT_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static UNLINKAT_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[expect(clippy::disallowed_methods)]
+pub(crate) static LINKAT_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: pipe2(2) $arg2
+#[expect(clippy::disallowed_methods)]
+pub static PIPE2_COOKIE_ARG2: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: pipe2(2) $arg3
+#[expect(clippy::disallowed_methods)]
+pub static PIPE2_COOKIE_ARG3: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: pipe2(2) $arg4
+#[expect(clippy::disallowed_methods)]
+pub static PIPE2_COOKIE_ARG4: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: pipe2(2) $arg5
+#[expect(clippy::disallowed_methods)]
+pub static PIPE2_COOKIE_ARG5: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+
 /// These are used in `syd::fs::seccomp_notify_addfd`.
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 
 /// These are used in `syd::fs::seccomp_notify_respond`.
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG3: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 
 /// These are used in `syd::proc::procmap_query`.
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static PROCMAP_QUERY_COOKIE_ARG3: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static PROCMAP_QUERY_COOKIE_ARG4: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
-#[allow(clippy::disallowed_methods)]
+#[expect(clippy::disallowed_methods)]
 pub(crate) static PROCMAP_QUERY_COOKIE_ARG5: LazyLock<Cookie> =
     LazyLock::new(|| getcookie().expect("getcookie"));
 
-/// Safe openat2 confined by syscall cookies.
+/// Safe openat2(2) confined by syscall cookies.
 #[inline(always)]
 pub(crate) fn safe_openat2<Fd: AsFd>(
     dirfd: Fd,
@@ -213,7 +252,7 @@ pub(crate) fn safe_openat2<Fd: AsFd>(
     mut how: OpenHow,
 ) -> Result<OwnedFd, Errno> {
     // SAFETY: In libc we trust.
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let fd = path.with_nix_path(|cstr| unsafe {
         libc::syscall(
             libc::SYS_openat2,
@@ -235,7 +274,7 @@ pub(crate) fn safe_openat2<Fd: AsFd>(
 }
 
 /// socket(2) may be multiplexed by socketcall(2).
-pub(crate) static SYS_SOCKET: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+pub static SYS_SOCKET: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("socket")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -256,7 +295,7 @@ pub fn safe_socket(
 ) -> Result<OwnedFd, Errno> {
     if let Some(sys_socket) = *SYS_SOCKET {
         // SAFETY: In libc we trust.
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         Errno::result(unsafe {
             libc::syscall(
                 sys_socket,
@@ -282,7 +321,7 @@ pub fn safe_socket(
 }
 
 /// accept4(2) may be multiplexed by socketcall(2).
-pub(crate) static SYS_ACCEPT4: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+pub static SYS_ACCEPT4: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("accept4")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -308,7 +347,7 @@ pub unsafe fn safe_accept4<Fd: AsFd>(
 ) -> Result<OwnedFd, Errno> {
     if let Some(sys_accept4) = *SYS_ACCEPT4 {
         // SAFETY: In libc we trust.
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         Errno::result(unsafe {
             libc::syscall(
                 sys_accept4,
@@ -334,6 +373,16 @@ pub unsafe fn safe_accept4<Fd: AsFd>(
     }
 }
 
+/// Safe accept4(2) confined by syscall cookies.
+///
+/// Unlike `safe_accept4`, this version does not return source address,
+/// therefore does not need an unsafe clause to dereference pointers.
+#[inline(always)]
+pub fn safer_accept4<Fd: AsFd>(fd: Fd, flags: SockFlag) -> Result<OwnedFd, Errno> {
+    // SAFETY: accept4(2) does not dereference NULL pointers.
+    unsafe { safe_accept4(fd, std::ptr::null_mut(), std::ptr::null_mut(), flags) }
+}
+
 /// bind(2) may be multiplexed by socketcall(2).
 pub(crate) static SYS_BIND: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("bind")
@@ -352,7 +401,6 @@ pub(crate) static SYS_BIND: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
 pub fn safe_bind<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno> {
     if let Some(sys_bind) = *SYS_BIND {
         // SAFETY: In libc we trust.
-        #[allow(clippy::cast_possible_truncation)]
         Errno::result(unsafe {
             libc::syscall(
                 sys_bind,
@@ -372,7 +420,7 @@ pub fn safe_bind<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno>
 }
 
 /// connect(2) may be multiplexed by socketcall(2).
-pub(crate) static SYS_CONNECT: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+pub static SYS_CONNECT: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("connect")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -389,7 +437,6 @@ pub(crate) static SYS_CONNECT: LazyLock<Option<libc::c_long>> = LazyLock::new(||
 pub fn safe_connect<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno> {
     if let Some(sys_connect) = *SYS_CONNECT {
         // SAFETY: In libc we trust.
-        #[allow(clippy::cast_possible_truncation)]
         Errno::result(unsafe {
             libc::syscall(
                 sys_connect,
@@ -408,11 +455,55 @@ pub fn safe_connect<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Err
     }
 }
 
-/// Safe memfd_create confined by syscall cookies.
+/// shutdown(2) may be multiplexed by socketcall(2).
+///
+/// This is used by syd-tor(1) only, syd(1) does not hook into shutdown(2).
+pub static SYS_SHUTDOWN: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+    match ScmpSyscall::from_name("shutdown")
+        .map(i32::from)
+        .map(libc::c_long::from)
+        .ok()
+    {
+        Some(n) if n < 0 => None,
+        Some(n) => Some(n),
+        None => None,
+    }
+});
+
+/// Safe shutdown(2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_shutdown<Fd: AsFd>(fd: Fd, how: Shutdown) -> Result<(), Errno> {
+    let how = match how {
+        Shutdown::Read => libc::SHUT_RD,
+        Shutdown::Write => libc::SHUT_WR,
+        Shutdown::Both => libc::SHUT_RDWR,
+    };
+
+    if let Some(sys_shutdown) = *SYS_SHUTDOWN {
+        // SAFETY: In libc we trust.
+        Errno::result(unsafe {
+            libc::syscall(
+                sys_shutdown,
+                fd.as_fd().as_raw_fd(),
+                how,
+                *SHUTDOWN_COOKIE_ARG2,
+                *SHUTDOWN_COOKIE_ARG3,
+                *SHUTDOWN_COOKIE_ARG4,
+                *SHUTDOWN_COOKIE_ARG5,
+            )
+        })
+        .map(drop)
+    } else {
+        // SAFETY: socketcall(2) on multiplexed architecture.
+        Errno::result(unsafe { libc::shutdown(fd.as_fd().as_raw_fd(), how) }).map(drop)
+    }
+}
+
+/// Safe memfd_create(2) confined by syscall cookies.
 #[inline(always)]
 pub(crate) fn safe_memfd_create(name: &[u8], flags: libc::c_uint) -> Result<OwnedFd, Errno> {
     // SAFETY: In libc we trust.
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let fd = Errno::result(unsafe {
         libc::syscall(
             libc::SYS_memfd_create,
@@ -513,6 +604,11 @@ static SYS_FTRUNCATE64: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
 
 /// Safe truncate(2) confined by syscall cookies.
 pub(crate) fn safe_truncate(path: &XPath, len: libc::off_t) -> Result<(), Errno> {
+    // On ILP32 where off_t is 64-bit, dispatch to truncate64 ABI.
+    if size_of::<libc::off_t>() > size_of::<libc::c_long>() {
+        return safe_truncate64(path, libc::off64_t::from(len));
+    }
+
     let sys_truncate = SYS_TRUNCATE.ok_or(Errno::ENOSYS)?;
 
     // SAFETY: In libc we trust.
@@ -611,6 +707,11 @@ pub(crate) fn safe_truncate64(path: &XPath, len: libc::off64_t) -> Result<(), Er
 
 /// Safe ftruncate(2) confined by syscall cookies.
 pub(crate) fn safe_ftruncate<Fd: AsFd>(fd: Fd, len: libc::off_t) -> Result<(), Errno> {
+    // On ILP32 where off_t is 64-bit, dispatch to ftruncate64 ABI.
+    if size_of::<libc::off_t>() > size_of::<libc::c_long>() {
+        return safe_ftruncate64(fd, libc::off64_t::from(len));
+    }
+
     let sys_ftruncate = SYS_FTRUNCATE.ok_or(Errno::ENOSYS)?;
 
     // SAFETY: In libc we trust.
@@ -760,3 +861,47 @@ pub(crate) fn safe_linkat<Fd1: AsFd, Fd2: AsFd>(
 
     Errno::result(res).map(drop)
 }
+
+/// Safe uname(2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_uname() -> Result<UtsName, Errno> {
+    let mut name = UtsName::default();
+
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        libc::syscall(
+            libc::SYS_uname,
+            &raw mut name,
+            *UNAME_COOKIE_ARG1,
+            *UNAME_COOKIE_ARG2,
+            *UNAME_COOKIE_ARG3,
+            *UNAME_COOKIE_ARG4,
+            *UNAME_COOKIE_ARG5,
+        )
+    })?;
+
+    Ok(name)
+}
+
+/// Safe pipe2(2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_pipe2(flags: OFlag) -> Result<(OwnedFd, OwnedFd), Errno> {
+    let mut fds = MaybeUninit::<[OwnedFd; 2]>::uninit();
+
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe {
+        libc::syscall(
+            libc::SYS_pipe2,
+            fds.as_mut_ptr(),
+            flags.bits(),
+            *PIPE2_COOKIE_ARG2,
+            *PIPE2_COOKIE_ARG3,
+            *PIPE2_COOKIE_ARG4,
+            *PIPE2_COOKIE_ARG5,
+        )
+    })?;
+
+    // SAFETY: pipe2 returns valid FDs on success.
+    let [read, write] = unsafe { fds.assume_init() };
+    Ok((read, write))
+}
