@@ -137,11 +137,11 @@ fn do_accept4<Fd: AsFd>(
     // Allocate storage for the address.
     let mut addr: [u8; std::mem::size_of::<SockaddrStorage>()] =
         [0u8; std::mem::size_of::<SockaddrStorage>()];
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let mut len = std::mem::size_of::<SockaddrStorage>() as libc::socklen_t;
 
     // Cast the storage buffer to a sockaddr pointer.
-    #[allow(clippy::cast_ptr_alignment)]
+    #[expect(clippy::cast_ptr_alignment)]
     let ptr = addr.as_mut_ptr() as *mut libc::sockaddr;
 
     // SAFETY: Record blocking call so it can get invalidated.
