@@ -26,13 +26,13 @@ use crate::{
     cookie::safe_bind,
     fs::CanonicalPath,
     hook::UNotifyEventRequest,
+    info,
     path::XPathBuf,
     proc::proc_umask,
     sandbox::{Action, AddressPattern, Capability, CidrRule},
     wildmatch::MatchMethod,
 };
 
-#[allow(clippy::cognitive_complexity)]
 pub(crate) fn handle_bind(
     fd: OwnedFd,
     addr: &SockaddrStorage,
@@ -48,7 +48,7 @@ pub(crate) fn handle_bind(
         // 1. Honour directory for too long sockets.
         //    Note, the current working directory is per-thread here.
         // 2. We cannot resolve symlinks in root or we risk TOCTOU!
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         let dirfd = root.as_ref().unwrap().dir.as_ref().unwrap();
         fchdir(dirfd)?;
 
@@ -63,6 +63,27 @@ pub(crate) fn handle_bind(
         // attempt to create the symlink target. Hence basename
         // in addr here is not vulnerable to TOCTOU.
         safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_sockaddr_in6() {
+        // SAFETY: addr is an IPV6 socket.
+        safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_sockaddr_in() {
+        // SAFETY: addr is an IPV4 socket.
+        safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_alg_addr() {
+        // SAFETY: addr is an Algorithm socket.
+        safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_link_addr() {
+        // SAFETY: addr is a Link socket.
+        safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_netlink_addr() {
+        // SAFETY: addr is a Netlink socket.
+        safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_vsock_addr() {
+        // SAFETY: addr is a Vsock socket.
+        safe_bind(&fd, addr)?;
+    } else if let Some(addr) = addr.as_unix_addr() {
+        // SAFETY: addr is a UNIX abstract socket.
+        safe_bind(&fd, addr)?;
     } else {
         // SAFETY: addr is not a UNIX domain socket.
         safe_bind(&fd, addr)?;
@@ -74,6 +95,7 @@ pub(crate) fn handle_bind(
     // Configure sandbox:
     // Remove and re-add the address so repeated binds to the
     // same address cannot overflow the vector.
+    #[expect(clippy::cognitive_complexity)]
     let _result = (|fd: OwnedFd, request: &UNotifyEventRequest| -> Result<(), Errno> {
         let (addr, port) = match addr.family() {
             Some(AddressFamily::Unix) => {
@@ -81,15 +103,15 @@ pub(crate) fn handle_bind(
                 let unix = match (addr.path(), addr.as_abstract()) {
                     (Some(_), _) => {
                         // Case 1: UNIX domain socket
-
+                        //
                         // SAFETY: addr.path()=Some asserts root is Some.
-                        #[allow(clippy::disallowed_methods)]
+                        #[expect(clippy::disallowed_methods)]
                         let unix = root.unwrap().abs();
 
                         // Handle bind_map after successful bind for UNIX sockets.
                         // We ignore errors because there's nothing we can do
                         // about them.
-                        let _ = request.add_bind(&fd, unix);
+                        let _ = request.add_unix(&fd, request.scmpreq.pid(), Some(unix));
                         drop(fd); // Close our copy of the socket.
 
                         if !allow_safe_bind {
@@ -100,25 +122,43 @@ pub(crate) fn handle_bind(
                     }
                     (_, Some(path)) => {
                         // Case 2: UNIX abstract socket
+                        //
+                        // SAFETY: Prefix UNIX abstract sockets with `@' before access check.
+                        let mut unix = XPathBuf::from("@");
+                        let null = memchr::memchr(0, path).unwrap_or(path.len());
+                        unix.append_bytes(&path[..null]);
 
+                        // Handle bind_map after successful bind for UNIX sockets.
+                        // We ignore errors because there's nothing we can do
+                        // about them.
+                        // BindMap is only used for SO_PEERCRED for UNIX abstract sockets.
+                        let _ = request.add_unix(&fd, request.scmpreq.pid(), None);
                         drop(fd); // Close our copy of the socket.
 
                         if !allow_safe_bind {
                             return Ok(());
                         }
 
-                        // SAFETY: Prefix UNIX abstract sockets with `@' before access check.
-                        let mut unix = XPathBuf::from("@");
-                        let null = memchr::memchr(0, path).unwrap_or(path.len());
-                        unix.append_bytes(&path[..null]);
-
                         unix
                     }
                     _ => {
                         // Case 3: unnamed UNIX socket.
                         //
                         // SAFETY: Use dummy path `!unnamed' for unnamed UNIX sockets.
-                        XPathBuf::from("!unnamed")
+                        let unix = XPathBuf::from("!unnamed");
+
+                        // Handle bind_map after successful bind for UNIX sockets.
+                        // We ignore errors because there's nothing we can do
+                        // about them.
+                        // BindMap is only used for SO_PEERCRED for UNIX abstract sockets.
+                        let _ = request.add_unix(&fd, request.scmpreq.pid(), None);
+                        drop(fd); // Close our copy of the socket.
+
+                        if !allow_safe_bind {
+                            return Ok(());
+                        }
+
+                        unix
                     }
                 };
 
@@ -138,8 +178,8 @@ pub(crate) fn handle_bind(
 
                 let addr = addr.as_sockaddr_in().ok_or(Errno::EINVAL)?;
                 let mut port = addr.port();
-
                 let addr = IpNet::new_assert(IpAddr::V4(addr.ip()), 32);
+
                 if port == 0 {
                     port = getsockname::<SockaddrStorage>(fd.as_raw_fd())?
                         .as_sockaddr_in()
@@ -157,8 +197,13 @@ pub(crate) fn handle_bind(
 
                 let addr = addr.as_sockaddr_in6().ok_or(Errno::EINVAL)?;
                 let mut port = addr.port();
+                let addr = addr.ip();
+                let addr = if let Some(addr) = addr.to_ipv4_mapped() {
+                    IpNet::new_assert(IpAddr::V4(addr), 32)
+                } else {
+                    IpNet::new_assert(IpAddr::V6(addr), 128)
+                };
 
-                let addr = IpNet::new_assert(IpAddr::V6(addr.ip()), 128);
                 if port == 0 {
                     port = getsockname::<SockaddrStorage>(fd.as_raw_fd())?
                         .as_sockaddr_in6()
@@ -172,13 +217,18 @@ pub(crate) fn handle_bind(
             _ => return Ok(()),
         };
 
+        let addr = AddressPattern {
+            addr,
+            port: Some(port..=port),
+        };
+        info!("ctx": "bind", "op": "allow_safe_bind",
+            "sys": "bind", "pid": request.scmpreq.pid().as_raw(), "rule": &addr,
+            "msg": format!("add rule `allow/net/connect+{addr}' after bind"));
+
         let rule = CidrRule {
             act: Action::Allow,
             cap: Capability::CAP_NET_CONNECT,
-            pat: AddressPattern {
-                addr,
-                port: Some(port..=port),
-            },
+            pat: addr,
         };
 
         let mut sandbox = request.get_mut_sandbox();
