@@ -12,17 +12,16 @@ use std::{
     net::IpAddr,
     ops::Deref,
     os::{
-        fd::{AsFd, AsRawFd, RawFd},
+        fd::{AsRawFd, RawFd},
         unix::ffi::OsStrExt,
     },
 };
 
-use ipnet::IpNet;
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
-    sys::socket::{getsockname, AddressFamily, MsgFlags, SockaddrLike, SockaddrStorage},
+    sys::socket::{AddressFamily, MsgFlags, SockaddrLike, SockaddrStorage},
     NixPath,
 };
 
@@ -42,7 +41,7 @@ use crate::{
         socket::handle_socket,
     },
     path::{XPath, XPathBuf},
-    sandbox::{Action, AddressPattern, Capability, CidrRule, SandboxGuard},
+    sandbox::{Action, Capability, SandboxGuard},
     warn,
 };
 
@@ -57,7 +56,6 @@ pub(crate) mod socket;
 
 const UNIX_PATH_MAX: usize = 108;
 
-#[allow(clippy::cognitive_complexity)]
 pub(crate) fn sys_socketcall(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
@@ -67,14 +65,15 @@ pub(crate) fn sys_socketcall(request: UNotifyEventRequest) -> ScmpNotifResp {
     // 0x3 => connect
     // 0x5 => accept
     // 0x6 => getsockname
+    // 0x9 => send
     // 0xb => sendto
     // 0xf => getsockopt
     // 0x10 => sendmsg
     // 0x12 => accept4
     // 0x14 => sendmmsg
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     let op = match req.data.args[0] {
-        n @ (0x1 | 0x2 | 0x3 | 0x5 | 0x6 | 0xb | 0xf | 0x10 | 0x12 | 0x14) => n as u8,
+        n @ (0x1 | 0x2 | 0x3 | 0x5 | 0x6 | 0x9 | 0xb | 0xf | 0x10 | 0x12 | 0x14) => n as u8,
         _ => {
             // SAFETY: Safe network call, continue.
             // No pointer-dereference in access check.
@@ -88,7 +87,7 @@ pub(crate) fn sys_socketcall(request: UNotifyEventRequest) -> ScmpNotifResp {
     let sizeof_ulong: usize = if is32 { 4 } else { 8 };
     const ARGLEN: usize = 6;
     let mut args = [0u64; ARGLEN];
-    #[allow(clippy::arithmetic_side_effects)]
+    #[expect(clippy::arithmetic_side_effects)]
     let bufsiz = sizeof_ulong * ARGLEN;
     let mut buf = Vec::new();
     if buf.try_reserve(bufsiz).is_err() {
@@ -175,7 +174,7 @@ pub(crate) fn sys_sendmmsg(request: UNotifyEventRequest) -> ScmpNotifResp {
 }
 
 // A helper function to handle network-related syscalls.
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8) -> ScmpNotifResp {
     syscall_handler!(request, |request: UNotifyEventRequest| {
         let sandbox = request.get_sandbox();
@@ -203,7 +202,7 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
                 Capability::empty()
             }
             0x2 /* bind */ => Capability::CAP_NET_BIND,
-            _ /* connect, send{to,{m,}msg} */ => Capability::CAP_NET_CONNECT,
+            _ /* connect, send{,to,{m,}msg} */ => Capability::CAP_NET_CONNECT,
         };
         drop(sandbox); // release the read-lock before get-fd.
 
@@ -213,7 +212,7 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
         // descriptors without having to resort to access()'ing
         // /proc/$pid/fd/$fd which will return ENOENT with
         // /proc mounted as hidepid=2.
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         let fd = request.get_fd(args[0] as RawFd)?;
 
         // SAFETY:
@@ -235,6 +234,12 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
                 // getsockname is used for informational purposes only.
                 return handle_getsockname(fd, &request, args);
             }
+            0x9 => {
+                // send: Connection mode socket.
+                // This only happens via socketcall(2) multiplexer.
+                // On 64-bit we do not hook into send(2).
+                return handle_sendto(fd, args, &request, None, restrict_oob);
+            }
             0xf => {
                 // getsockopt is used for informational purposes only.
                 return handle_getsockopt(fd, &request, args);
@@ -266,8 +271,8 @@ fn syscall_network_handler(request: UNotifyEventRequest, args: &[u64; 6], op: u8
 
         let idx = if op == 0xb /* sendto */ { 4 } else { 1 };
         let addr_remote = args[idx];
-        #[allow(clippy::arithmetic_side_effects)]
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::arithmetic_side_effects)]
+        #[expect(clippy::cast_possible_truncation)]
         let addr_len = args[idx + 1] as libc::socklen_t;
         if addr_remote == 0 && addr_len == 0 {
             if op == 0xb {
@@ -341,7 +346,7 @@ fn get_addr(
     if addr_len < 3 {
         return Err(Errno::EINVAL);
     }
-    #[allow(clippy::arithmetic_side_effects)]
+    #[expect(clippy::arithmetic_side_effects)]
     let addr_len = addr_len.min(std::mem::size_of::<libc::sockaddr_un>() + UNIX_PATH_MAX);
 
     let mut addr = Vec::new();
@@ -356,10 +361,7 @@ fn get_addr(
     // 1. The memory location of `sockaddr_ptr` is valid, correctly aligned.
     // 2. The memory is allocated based on a valid `sockaddr` structure.
     // 3. There are no concurrent writes to the memory location while reading.
-    match unsafe {
-        #[allow(clippy::cast_ptr_alignment)]
-        SockaddrStorage::from_raw(addr, Some(addr_len))
-    } {
+    match unsafe { SockaddrStorage::from_raw(addr, Some(addr_len)) } {
         Some(addr)
             if addr.as_sockaddr_in().is_some()
                 && (addr_len as usize) < std::mem::size_of::<libc::sockaddr_in>() =>
@@ -389,7 +391,7 @@ fn canon_addr<'a>(
     addr: SockaddrStorage,
     cap: Capability,
 ) -> Result<(SockaddrStorage, Option<CanonicalPath<'a>>), Errno> {
-    #[allow(clippy::cast_possible_truncation)]
+    #[expect(clippy::cast_possible_truncation)]
     if let Some(path) = addr.as_unix_addr().and_then(|a| a.path()) {
         // Check for chroot.
         if sandbox.is_chroot() {
@@ -423,7 +425,7 @@ fn canon_addr<'a>(
             // proc path /proc/thread-self/fd/$fd in address' path
             // argument to avoid symlink TOCTOU because connect and
             // sendto follow symlinks in basename unlike bind.
-            #[allow(clippy::disallowed_methods)]
+            #[expect(clippy::disallowed_methods)]
             let fd = path.dir.as_ref().unwrap();
             let mut pfd = XPathBuf::from("/proc/thread-self/fd");
             pfd.push_fd(fd.as_raw_fd());
@@ -462,7 +464,7 @@ fn canon_addr<'a>(
         // including the family and the path. The size is the offset of
         // the sun_path field plus the length of the path (including the
         // null terminator).
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         let size = std::mem::size_of::<libc::sa_family_t>() + socklen;
 
         // SAFETY: We are converting a sockaddr_un to a
@@ -537,7 +539,7 @@ pub(crate) fn sandbox_addr(
 }
 
 /// Process a `AddressFamily::Unix` socket address.
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub(crate) fn sandbox_addr_unix(
     request: &UNotifyEventRequest,
     sandbox: &SandboxGuard,
@@ -634,7 +636,7 @@ pub(crate) fn sandbox_addr_unix(
 }
 
 /// Process an IPv4 or IPv6 address.
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub(crate) fn sandbox_addr_inet(
     request: &UNotifyEventRequest,
     sandbox: &SandboxGuard,
@@ -669,13 +671,7 @@ pub(crate) fn sandbox_addr_inet(
     let (addr, port) = if let Some(sin) = addr.as_sockaddr_in() {
         (IpAddr::V4(sin.ip()), sin.port())
     } else if let Some(sa6) = addr.as_sockaddr_in6() {
-        if let Some(v4addr) = sa6.ip().to_ipv4_mapped() {
-            // Mapped IPv4 address, convert to IPv4.
-            (IpAddr::V4(v4addr), sa6.port())
-        } else {
-            // Regular IPv6 address, check as is.
-            (IpAddr::V6(sa6.ip()), sa6.port())
-        }
+        (sa6.ip().to_canonical(), sa6.port())
     } else {
         return Err(Errno::EINVAL);
     };
@@ -743,51 +739,3 @@ pub(crate) fn to_msgflags(arg: u64) -> Result<MsgFlags, Errno> {
     let flags = arg.try_into().or(Err(Errno::EINVAL))?;
     MsgFlags::from_bits(flags).ok_or(Errno::EINVAL)
 }
-
-// Handle allow_safe_bind for connect and sendto.
-// bind is handled privately in handle_bind.
-pub(crate) fn handle_safe_bind<Fd: AsFd>(
-    request: &UNotifyEventRequest,
-    fd: Fd,
-) -> Result<(), Errno> {
-    let addr = getsockname::<SockaddrStorage>(fd.as_fd().as_raw_fd())?;
-
-    let (addr, port) = if let Some(addr) = addr.as_sockaddr_in() {
-        let port = addr.port();
-        if port == 0 {
-            return Ok(());
-        }
-
-        // Allow implicit bind with safe_bind.
-        (IpNet::new_assert(IpAddr::V4(addr.ip()), 32), port)
-    } else if let Some(addr) = addr.as_sockaddr_in6() {
-        let port = addr.port();
-        if port == 0 {
-            return Ok(());
-        }
-
-        // Allow implicit bind with safe_bind.
-        (IpNet::new_assert(IpAddr::V6(addr.ip()), 128), port)
-    } else {
-        return Ok(());
-    };
-
-    // Configure sandbox:
-    // Remove and re-add the address so repeated binds to the
-    // same address cannot overflow the vector.
-    let rule = CidrRule {
-        act: Action::Allow,
-        cap: Capability::CAP_NET_CONNECT,
-        pat: AddressPattern {
-            addr,
-            port: Some(port..=port),
-        },
-    };
-    let mut sandbox = request.get_mut_sandbox();
-    if let Some(idx) = sandbox.cidr_rules.iter().position(|r| *r == rule) {
-        sandbox.cidr_rules.remove(idx);
-    }
-    sandbox.cidr_rules.push_front(rule)?;
-
-    Ok(())
-}
