@@ -34,8 +34,7 @@ use crate::{
     sandbox::Capability,
 };
 
-#[allow(clippy::cognitive_complexity)]
-#[allow(clippy::too_many_arguments)]
+#[expect(clippy::cognitive_complexity)]
 pub(crate) fn handle_sendmsg(
     fd: OwnedFd,
     request: &UNotifyEventRequest,
@@ -129,10 +128,9 @@ pub(crate) fn handle_sendmsg(
     };
 
     // Step 2: Handle control messages.
-    #[allow(clippy::collection_is_never_read)]
     let mut control_messages = Vec::new();
     let control_data = if !msg.msg_control.is_null() && msg.msg_controllen > 0 {
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::useless_conversion)]
         let cmsg_len = usize::try_from(msg.msg_controllen)
             .or(Err(Errno::EINVAL))?
             .min(1000000); // SAFETY: Cap at 1mio.
@@ -267,7 +265,7 @@ pub(crate) fn handle_sendmsg(
         // SAFETY: The msg_iovlen member of the msghdr struct
         // must not be fully trusted, it can be overly large,
         // and allocating a Vector of that capacity may overflow.
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::useless_conversion)]
         let len = usize::try_from(msg.msg_iovlen)
             .or(Err(Errno::EINVAL))?
             .min(1000000); // Cap count at 1mio
@@ -351,7 +349,6 @@ pub(crate) fn handle_sendmsg(
         false
     };
 
-    #[allow(clippy::cast_possible_wrap)]
     let result = if let Some((addr, _)) = addr {
         // UNIX domain/abstract socket.
         sendmsg(
@@ -364,9 +361,7 @@ pub(crate) fn handle_sendmsg(
     } else {
         // Connection-mode socket.
         sendmsg::<SockaddrStorage>(fd.as_raw_fd(), &io_slices, &control_messages, flags, None)
-    }
-    .map(|n| request.return_syscall(n as i64));
-    drop(fd); // Close our copy of the socket.
+    };
 
     // Remove invalidation record unless interrupted.
     if is_blocking {
@@ -375,11 +370,19 @@ pub(crate) fn handle_sendmsg(
             .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
     }
 
-    result
+    // Send SIGPIPE for EPIPE unless MSG_NOSIGNAL is set.
+    #[expect(clippy::cast_possible_wrap)]
+    Ok(match result {
+        Ok(n) => request.return_syscall(n as i64),
+        Err(Errno::EPIPE) if !flags.contains(MsgFlags::MSG_NOSIGNAL) => {
+            request.pidfd_kill(libc::SIGPIPE)?;
+            request.fail_syscall(Errno::EPIPE)
+        }
+        Err(errno) => request.fail_syscall(errno),
+    })
 }
 
-#[allow(clippy::cognitive_complexity)]
-#[allow(clippy::too_many_arguments)]
+#[expect(clippy::cognitive_complexity)]
 pub(crate) fn handle_sendmmsg(
     fd: OwnedFd,
     request: &UNotifyEventRequest,
@@ -447,6 +450,8 @@ pub(crate) fn handle_sendmmsg(
         } else {
             (false, false)
         };
+    // Check if we want to send SIGPIPE on EPIPE.
+    let must_signal = !msgflags.contains(MsgFlags::MSG_NOSIGNAL);
 
     // Prepare a series of sendmsg calls.
     for mmsg in &mut mmsghdrs {
@@ -496,7 +501,7 @@ pub(crate) fn handle_sendmmsg(
         // Step 2: Handle control messages.
         let mut control_data = Vec::new();
         let control_datum = if !msg.msg_control.is_null() && msg.msg_controllen > 0 {
-            #[allow(clippy::useless_conversion)]
+            #[expect(clippy::useless_conversion)]
             let cmsg_len = usize::try_from(msg.msg_controllen)
                 .or(Err(Errno::EINVAL))?
                 .min(1000000); // SAFETY: Cap at 1mio.
@@ -637,7 +642,7 @@ pub(crate) fn handle_sendmmsg(
             // SAFETY: The msg_iovlen member of the msghdr struct
             // must not be fully trusted, it can be overly large,
             // and allocating a Vector of that capacity may overflow.
-            #[allow(clippy::useless_conversion)]
+            #[expect(clippy::useless_conversion)]
             let len = usize::try_from(msg.msg_iovlen)
                 .or(Err(Errno::EINVAL))?
                 .min(1000000); // Cap count at 1mio.
@@ -727,10 +732,19 @@ pub(crate) fn handle_sendmmsg(
                     .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
             }
 
+            // Send SIGPIPE for EPIPE unless MSG_NOSIGNAL is set.
+            mmsg.msg_len = match result {
+                Ok(n) => n.try_into().or(Err(Errno::EINVAL))?,
+                Err(Errno::EPIPE) if must_signal => {
+                    request.pidfd_kill(libc::SIGPIPE)?;
+                    return Err(Errno::EPIPE);
+                }
+                Err(errno) => return Err(errno),
+            };
+
             mmsg.msg_len = result?.try_into().or(Err(Errno::EINVAL))?;
         }
     }
-    drop(fd); // Close our copy of the socket.
 
     // Write back mmsghdr structures to remote process memory
     let mut buf: Zeroizing<Vec<u8>> = Zeroizing::new(Vec::new());
@@ -756,12 +770,12 @@ pub(crate) fn handle_sendmmsg(
     request.write_mem(&buf, addr)?;
 
     // FIXME: We do not handle partial success.
-    #[allow(clippy::cast_possible_wrap)]
+    #[expect(clippy::cast_possible_wrap)]
     Ok(request.return_syscall(mmsghdrs.len() as i64))
 }
 
 // SAFETY: Below lie daemons...
-#[allow(clippy::type_complexity)]
+#[expect(clippy::type_complexity)]
 fn parse_control_messages(
     request: &UNotifyEventRequest,
     cmsg_buf: &[u8],
@@ -806,7 +820,6 @@ fn parse_control_messages(
         // the buffer is initially populated in a manner adhering to the alignment
         // requirements of `cmsghdr`.  The caller is responsible for ensuring that `offset`
         // is correctly aligned for `cmsghdr` when accessing the buffer.
-        #[allow(clippy::cast_ptr_alignment)]
         let (cmsg_header, cmsg_len0): (crate::compat::cmsghdr, usize) = if is32 {
             // SAFETY: See the comment above.
             let cmsg_header_32: crate::compat::cmsghdr32 = unsafe {
@@ -830,7 +843,7 @@ fn parse_control_messages(
         if cmsg_header.cmsg_len < cmsg_len0 {
             return Err(Errno::EINVAL); // Invalid header length
         }
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::useless_conversion)]
         let data_len: usize = cmsg_header.cmsg_len.try_into().or(Err(Errno::EINVAL))?;
         let data_len = data_len.checked_sub(cmsg_len0).ok_or(Errno::EINVAL)?;
 
@@ -862,9 +875,9 @@ fn parse_control_messages(
                 // system for `SCM_RIGHTS` control messages, ensuring that we read valid
                 // file descriptor values from the buffer.
                 for i in 0..fd_count {
-                    #[allow(clippy::arithmetic_side_effects)]
+                    #[expect(clippy::arithmetic_side_effects)]
                     let fd_offset = i * size_of::<RawFd>();
-                    #[allow(clippy::cast_ptr_alignment)]
+                    #[expect(clippy::cast_ptr_alignment)]
                     // SAFETY: See the comment above.
                     let fd = unsafe { *(data[fd_offset..].as_ptr() as *const RawFd) };
                     let fd = request.get_fd(fd)?;
@@ -873,7 +886,7 @@ fn parse_control_messages(
                 control_fds.push(fds);
             }
             (libc::SOL_SOCKET, libc::SCM_CREDENTIALS) => {
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: Casting `data.as_ptr()` to `*const libc::ucred` is safe
                 // under the assumption that `data` contains bytes that correctly represent
                 // a `libc::ucred` structure, and that `data_len` matches the size of
@@ -935,7 +948,7 @@ fn parse_control_messages(
                 if data_len != size_of::<libc::c_int>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let op = unsafe { *(data.as_ptr() as *const libc::c_int) };
                 control_ops.push(op); // Store the operation code
@@ -956,7 +969,7 @@ fn parse_control_messages(
                 if data_len != size_of::<u32>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let assoclen = unsafe { *(data.as_ptr() as *const u32) };
                 control_aead_assoclens.push(assoclen); // Store the AEAD assoclen
@@ -977,7 +990,7 @@ fn parse_control_messages(
                 if data_len != size_of::<u16>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let gso_segments = unsafe { *(data.as_ptr() as *const u16) };
                 control_udp_gso_segments.push(gso_segments); // Store the GSO segment count
@@ -997,7 +1010,7 @@ fn parse_control_messages(
                 if data_len != size_of::<libc::in_pktinfo>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let pktinfo = unsafe { &*(data.as_ptr() as *const libc::in_pktinfo) };
                 control_ipv4_packet_infos.push(*pktinfo); // Store the IPv4 packet info
@@ -1017,7 +1030,7 @@ fn parse_control_messages(
                 if data_len != size_of::<libc::in6_pktinfo>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let pktinfo = unsafe { &*(data.as_ptr() as *const libc::in6_pktinfo) };
                 control_ipv6_packet_infos.push(*pktinfo); // Store the IPv6 packet info
@@ -1038,7 +1051,7 @@ fn parse_control_messages(
                 if data_len != size_of::<u32>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let rxq_ovfl = unsafe { *(data.as_ptr() as *const u32) };
                 control_rxq_ovfls.push(rxq_ovfl); // Store the Rx queue overflow count
@@ -1058,7 +1071,7 @@ fn parse_control_messages(
                 if data_len != size_of::<u64>() {
                     return Err(Errno::EINVAL); // Data length mismatch
                 }
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 // SAFETY: See the comment above.
                 let tx_time = unsafe { *(data.as_ptr() as *const u64) };
                 control_tx_times.push(tx_time); // Store the Tx time
@@ -1070,7 +1083,7 @@ fn parse_control_messages(
                 }
                 // SAFETY: Exact int-sized payload validated; bounds and alignment guaranteed
                 // by prior checks and cmsg parsing semantics. Single c_int read is safe.
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 let val = unsafe { *(data.as_ptr() as *const libc::c_int) };
                 // See ip_cmd_send() in Linux' net/ipv4/ip_sockglue.c.
                 let val: u8 = val.try_into().or(Err(Errno::EINVAL))?;
@@ -1082,7 +1095,7 @@ fn parse_control_messages(
                 }
                 // SAFETY: Exact int-sized payload validated; bounds and alignment guaranteed
                 // by prior checks and cmsg parsing semantics. Single c_int read is safe.
-                #[allow(clippy::cast_ptr_alignment)]
+                #[expect(clippy::cast_ptr_alignment)]
                 let val = unsafe { *(data.as_ptr() as *const i32) };
                 control_ipv6_tclasses.push(val);
             }
@@ -1101,8 +1114,8 @@ fn parse_control_messages(
         // practice for parsing sequences of control messages in a buffer, provided that the
         // control message buffer (`cmsg_buf`) is initially aligned and structured correctly
         // according to control message protocols.
-        #[allow(clippy::arithmetic_side_effects)]
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::arithmetic_side_effects)]
+        #[expect(clippy::cast_possible_truncation)]
         if !is32 {
             // SAFETY: See the comment above.
             offset += unsafe { libc::CMSG_SPACE(data_len as u32) } as usize;
