@@ -12,6 +12,7 @@ use nix::{
     errno::Errno,
     sys::socket::{send, sendto, MsgFlags, SockaddrStorage},
 };
+use zeroize::Zeroizing;
 
 use crate::{
     fs::{get_nonblock, has_recv_timeout},
@@ -24,20 +25,34 @@ pub(crate) fn handle_sendto(
     args: &[u64; 6],
     request: &UNotifyEventRequest,
     addr: Option<&SockaddrStorage>,
+    restrict_oob: bool,
 ) -> Result<ScmpNotifResp, Errno> {
     // SAFETY: Reject undefined/invalid flags.
     let flags = to_msgflags(args[3])?;
 
-    // SAFETY: The length argument to the sendto call
-    // must not be fully trusted, it can be overly large,
-    // and allocating a Vector of that capacity may overflow.
+    // SAFETY: Reject MSG_OOB as necessary.
+    if restrict_oob && flags.contains(MsgFlags::MSG_OOB) {
+        // Signal no support to let the sandbox process
+        // handle the error gracefully. This is consistent
+        // with the Linux kernel.
+        return Err(Errno::EOPNOTSUPP);
+    }
+
+    // SAFETY:
+    // 1. The length argument to the sendto call
+    //    must not be fully trusted, it can be overly large,
+    //    and allocating a Vector of that capacity may overflow.
+    // 2. It is valid for the length to be zero to send an empty message.
+    // 3. Buffer read from sandbox process MUST be zeroized on drop.
     let len = usize::try_from(args[2])
         .or(Err(Errno::EINVAL))?
         .min(1000000); // Cap count at 1mio.
-    let mut buf = Vec::new();
-    buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-    buf.resize(len, 0);
-    request.read_mem(&mut buf, args[1])?;
+    let mut buf = Zeroizing::new(Vec::new());
+    if len > 0 {
+        buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
+        buf.resize(len, 0);
+        request.read_mem(&mut buf, args[1])?;
+    }
 
     // SAFETY: Record blocking call so it can get invalidated.
     let req = request.scmpreq;
@@ -52,15 +67,13 @@ pub(crate) fn handle_sendto(
         false
     };
 
-    #[allow(clippy::cast_possible_wrap)]
     let result = if let Some(addr) = addr {
         // Connection-less socket.
         sendto(fd.as_raw_fd(), &buf, addr, flags)
     } else {
         // Connection mode socket, no address specified.
         send(fd.as_raw_fd(), &buf, flags)
-    }
-    .map(|n| request.return_syscall(n as i64));
+    };
 
     // Remove invalidation record unless interrupted.
     if is_blocking {
@@ -69,5 +82,14 @@ pub(crate) fn handle_sendto(
             .del_sys_block(req.id, matches!(result, Err(Errno::EINTR)))?;
     }
 
-    result
+    // Send SIGPIPE for EPIPE unless MSG_NOSIGNAL is set.
+    #[expect(clippy::cast_possible_wrap)]
+    Ok(match result {
+        Ok(n) => request.return_syscall(n as i64),
+        Err(Errno::EPIPE) if !flags.contains(MsgFlags::MSG_NOSIGNAL) => {
+            request.pidfd_kill(libc::SIGPIPE)?;
+            request.fail_syscall(Errno::EPIPE)
+        }
+        Err(errno) => request.fail_syscall(errno),
+    })
 }
