@@ -25,11 +25,40 @@ use serde_json::{Map, Value};
 use crate::{
     config::*,
     err::SydResult,
-    fs::{is_active_fd, lock_fd, unlock_fd},
+    fs::{is_active_fd, lock_fd, retry_on_eintr, unlock_fd},
     proc::{proc_cmdline, proc_comm, proc_cwd, proc_tty},
     syslog::LogLevel,
 };
 
+/// Formatted print which returns Result as Errno rather than panicking.
+#[macro_export]
+macro_rules! printf {
+    () => {{}};
+    ($($arg:tt)*) => {{
+        use ::std::io::Write as _;
+        let __s = format!($($arg)*);
+        ::std::io::stdout().write_all(__s.as_bytes())
+            .map_err(|error| $crate::err::err2no(&error))
+    }};
+}
+
+/// Formatted println which returns Result as Errno rather than panicking.
+#[macro_export]
+macro_rules! printfln {
+    () => {{
+        use ::std::io::Write as _;
+        ::std::io::stdout().write_all(b"\n")
+            .map_err(|error| $crate::err::err2no(&error))
+    }};
+    ($($arg:tt)*) => {{
+        use ::std::io::Write as _;
+        let mut __s = format!($($arg)*);
+        __s.push('\n');
+        ::std::io::stdout().write_all(__s.as_bytes())
+            .map_err(|error| $crate::err::err2no(&error))
+    }};
+}
+
 // Whether we concluded the output is a TTY.
 pub(crate) static LOG_TTY: std::sync::atomic::AtomicBool =
     std::sync::atomic::AtomicBool::new(false);
@@ -51,7 +80,7 @@ pub(crate) fn log_set_panic_hook() {
     // Set a logging panic hook. The default panic
     // hook calls system calls not permitted by emulators
     // such as getcwd(2), stat(2) etc.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     std::panic::set_hook(Box::new(|info| {
         let this = std::thread::current();
         let name = this.name().unwrap_or("?");
@@ -68,7 +97,7 @@ pub(crate) fn log_set_panic_hook() {
         let line = info.location().map(|l| l.line());
 
         if log_is_main() {
-            // Main thread panicing isn't recoverable unlike others.
+            // Main thread panicking isn't recoverable unlike others.
             // Ensure clean exit right away.
             crate::alert!("ctx": "panic",
                 "op": "panic_syd_main_thread",
@@ -267,7 +296,7 @@ pub(crate) struct LockedWriter<'a> {
 
 impl<'a> LockedWriter<'a> {
     pub(crate) fn new(fd: BorrowedFd<'a>) -> Result<Self, Errno> {
-        lock_fd(fd, true, true)?;
+        retry_on_eintr(|| lock_fd(fd, true, true))?;
         Ok(Self { fd })
     }
 }
@@ -291,7 +320,7 @@ impl Write for LockedWriter<'_> {
 
 /// Initializes Syslog global object, reading environment variables.
 #[cfg(feature = "log")]
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub fn log_init(default_level: LogLevel, default_log_fd: Option<RawFd>) -> Result<(), Errno> {
     use std::os::unix::ffi::OsStrExt;
 
@@ -363,8 +392,13 @@ pub fn log_init(default_level: LogLevel, default_log_fd: Option<RawFd>) -> Resul
     init_global_syslog(logbuflen, fd, level, use_stack)?;
 
     // Finally let's make some noise!
-    info!("ctx": "init", "op": "sing", "chapter": 24,
-        "msg": "Change return success. Going and coming without error. Action brings good fortune.");
+    //
+    // ENV_SING avoids duplicating this message over memfd-reexec.
+    if std::env::var_os(ENV_SING).is_none() {
+        info!("ctx": "init", "op": "sing", "chapter": 24,
+            "msg": "Change return success. Going and coming without error. Action brings good fortune.");
+        std::env::set_var(ENV_SING, "");
+    }
 
     Ok(())
 }
@@ -436,7 +470,7 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
 
 /// Main entry point for appending log entries in JSON-line style.
 #[cfg(feature = "log")]
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String, Value>) {
     let sys = if let Some(sys) = crate::syslog::global_syslog() {
         sys
@@ -449,9 +483,9 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
     let tty = LOG_TTY.load(std::sync::atomic::Ordering::Relaxed);
 
     // If there's a "pid", we might add cmd/cwd, etc.
-    // We remove and readd to reorder for better visibility.
+    // We remove and re-add to reorder for better visibility.
     if let Some(pid_v) = msg.remove("pid").and_then(|v| v.as_i64()) {
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         let pid = Pid::from_raw(pid_v as libc::pid_t);
         if pid.as_raw() != 0 {
             if add_context {
@@ -531,7 +565,7 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
 /// Initializes Syslog global object, reading environment variables.
 #[cfg(not(feature = "log"))]
 #[inline(always)]
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub fn log_init(_default_level: LogLevel, default_log_fd: Option<RawFd>) -> Result<(), Errno> {
     use std::os::unix::ffi::OsStrExt;
 
@@ -596,8 +630,15 @@ pub fn log_init(_default_level: LogLevel, default_log_fd: Option<RawFd>) -> Resu
     if let Some(fd) = fd {
         LOG_FD.store(fd, std::sync::atomic::Ordering::Relaxed);
     }
-    info!("ctx": "init", "op": "sing", "chapter": 24,
-        "msg": "Change return success. Going and coming without error. Action brings good fortune.");
+
+    // Finally let's make some noise!
+    //
+    // ENV_SING avoids duplicating this message over memfd-reexec.
+    if std::env::var_os(ENV_SING).is_none() {
+        info!("ctx": "init", "op": "sing", "chapter": 24,
+            "msg": "Change return success. Going and coming without error. Action brings good fortune.");
+        std::env::set_var(ENV_SING, "");
+    }
 
     Ok(())
 }
@@ -613,7 +654,7 @@ pub fn log_init_simple(default_level: LogLevel) -> Result<(), Errno> {
 
 /// Main entry point for appending log entries in JSON-line style.
 #[cfg(not(feature = "log"))]
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String, Value>) {
     // Setting LOG-FD to negative is OK to disable logging.
     let fd = LOG_FD.load(std::sync::atomic::Ordering::Relaxed);
@@ -629,9 +670,9 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
     let tty = LOG_TTY.load(std::sync::atomic::Ordering::Relaxed);
 
     // If there's a "pid", we might add cmd/cwd, etc.
-    // We remove and readd to reorder for better visibility.
+    // We remove and re-add to reorder for better visibility.
     if let Some(pid_v) = msg.remove("pid").and_then(|v| v.as_i64()) {
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         let pid = Pid::from_raw(pid_v as libc::pid_t);
         if pid.as_raw() != 0 {
             if add_context {
@@ -730,18 +771,17 @@ pub fn now() -> u64 {
 /// The format is currently "YYYYMMDDThhmmssZ". The format may change in
 /// the future but it will always remain conformant to the ISO8601
 /// standard.
-#[allow(clippy::arithmetic_side_effects)]
-#[allow(deprecated)]
+#[expect(clippy::arithmetic_side_effects)]
 fn format_iso8601(timestamp: u64) -> String {
     let timestamp = match libc::time_t::try_from(timestamp) {
         Ok(t) => t,
-        Err(_) => return "error on time()".to_string(),
+        Err(_) => return "EOVERFLOW".to_string(),
     };
     let mut tm = std::mem::MaybeUninit::uninit();
 
-    // SAFETY: gmtime_r returns NULL on error.
-    if unsafe { libc::gmtime_r(&timestamp, tm.as_mut_ptr()) }.is_null() {
-        return "error on gmtime_r()".to_string();
+    // SAFETY: gmtime_r returns NULL on error and set errno(3).
+    if unsafe { libc::gmtime_r(&raw const timestamp, tm.as_mut_ptr()) }.is_null() {
+        return Errno::last().to_string();
     }
 
     // SAFETY: gmtime_r returned success.
