@@ -19,7 +19,7 @@ use nix::{
 };
 
 use crate::{
-    caps::{securebits::set_keepcaps, Capability},
+    caps::Capability,
     config::*,
     confine::{confine_scmp_ioctl, confine_scmp_pwritev2, safe_drop_cap},
     fs::retry_on_eintr,
@@ -68,7 +68,7 @@ macro_rules! fail_errno_safe {
     };
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
     // SAFETY: arg is a valid ChildInfo structure.
     let mut child: Box<ChildInfo> = unsafe { Box::from_raw(arg as *mut ChildInfo) };
@@ -275,7 +275,7 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         let fd = seccomp_fd.as_raw_fd().to_le_bytes();
         let mut nwrite = 0;
         while nwrite < fd.len() {
-            #[allow(clippy::arithmetic_side_effects)]
+            #[expect(clippy::arithmetic_side_effects)]
             match retry_on_eintr(|| write(&pipe_rw, &fd[nwrite..])) {
                 Ok(0) => {
                     // Parent died before reading.
@@ -354,19 +354,6 @@ pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
         fail_safe!(child, Err::CapSet);
     }
 
-    // Ensure we keep the specified capabilities.
-    if set_keepcaps(true).is_err() {
-        fail_safe!(child, Err::CapSet);
-    }
-
-    // TODO:
-    // Set the new secure bits:
-    // 1. SECBIT_EXEC_RESTRICT_FILE
-    // 2. SECBIT_EXEC_DENY_INTERACTIVE
-    // upon user configuration.
-    // See: https://docs.kernel.org/userspace-api/check_exec.html
-    // Note, we already use AT_EXECVE_CHECK.
-
     match child.exe {
         Exe::Library(lib) => unsafe {
             let fun: Symbol<unsafe extern "C" fn() -> i32> = match lib.get(b"syd_main") {
