@@ -18,6 +18,16 @@ use syd::{
     sandbox::Sandbox,
 };
 
+// Set global allocator to mimalloc.
+#[cfg(all(not(feature = "prof"), target_pointer_width = "64"))]
+#[global_allocator]
+static GLOBAL: mimalloc::MiMalloc = mimalloc::MiMalloc;
+
+// Set global allocator to tcmalloc if profiling is enabled.
+#[cfg(feature = "prof")]
+#[global_allocator]
+static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
+
 syd::main! {
     use lexopt::prelude::*;
 
@@ -91,7 +101,7 @@ syd::main! {
         let ips_ext = XPath::from_bytes(b"ipset");
         let net_ext = XPath::from_bytes(b"netset");
 
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         if *fext == *syd_ext {
             syd.parse_config_file(&path)?;
         } else if *fext == *ips_ext || *fext == *net_ext {
