@@ -473,9 +473,8 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     // Allow safe prctl(2) operations.
     let sysname = "prctl";
     if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-        for op in [libc::PR_SET_VMA as u64] {
-            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == op)])?;
-        }
+        let op = libc::PR_SET_VMA as u64;
+        ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == op)])?;
     }
 
     // Prevent executable memory.
