@@ -148,7 +148,7 @@ syd::main! {
                 let glob = sysarg.to_str().ok_or(Errno::EINVAL)?;
 
                 let mut ok = false;
-                for errno in (1..=4096).map(|n| Errno::from_raw(i32::from(n))) {
+                for errno in (1..=4096).map(Errno::from_raw) {
                     if errno == Errno::UnknownErrno {
                         continue;
                     }
