@@ -8,7 +8,6 @@
 
 use std::{
     borrow::Cow,
-    collections::HashSet,
     env,
     ffi::{CStr, CString, OsStr},
     fmt,
@@ -29,7 +28,7 @@ use memchr::memchr3;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
-    libc::{_exit, c_char, prctl, size_t, CLONE_FILES, ENOSYS, SIGCHLD, SIGKILL, SIGSYS},
+    libc::{_exit, c_char, size_t, CLONE_FILES, ENOSYS, SIGCHLD, SIGKILL, SIGSYS},
     mount::{mount, MsFlags},
     sched::{unshare, CloneFlags},
     sys::{
@@ -43,6 +42,7 @@ use nix::{
 use crate::{
     compat::{pipe2_raw, waitid, WaitStatus},
     config::{MINI_STACK_SIZE, *},
+    confine::{confine_mdwe, MS_NOSYMFOLLOW},
     debug,
     err::err2no,
     fs::{
@@ -50,12 +50,13 @@ use crate::{
         MFD_ALLOW_SEALING,
     },
     get_user_home, get_user_name,
+    hash::SydHashSet,
     landlock::RulesetStatus,
     landlock_policy::LandlockPolicy,
     log::contains_ascii_unprintable,
     path::PATH_MAX,
     unshare::{GidMap, UidMap},
-    xpath, XPath, XPathBuf, MS_NOSYMFOLLOW,
+    xpath, XPath, XPathBuf,
 };
 
 bitflags! {
@@ -144,7 +145,7 @@ impl From<i32> for WordExpError {
             // Used by pipe writer in the confined process.
             return Self::SystemError(Errno::from_raw(code));
         }
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         match code {
             WRDE_BADCHAR => Self::BadCharacter,
             WRDE_BADVAL => Self::BadValue,
@@ -160,7 +161,7 @@ impl From<i32> for WordExpError {
 
 impl From<WordExpError> for i32 {
     fn from(val: WordExpError) -> Self {
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         match val {
             WordExpError::BadCharacter => WRDE_BADCHAR,
             WordExpError::BadValue => WRDE_BADVAL,
@@ -291,7 +292,7 @@ impl WordExp<'_> {
     }
 
     /// Perform environment/tilde expansion and optionally command substitution.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub fn expand_full(input: &str, timeout: Duration) -> Result<Cow<'_, str>, WordExpError> {
         // Quick returns:
         // Empty string or no special characters present.
@@ -404,7 +405,7 @@ impl WordExp<'_> {
                     if let Err(ref error) = pipe.write_all(word.as_bytes()) {
                         let err = err2no(error) as i32;
                         // SAFETY: In libc we trust.
-                        #[allow(clippy::arithmetic_side_effects)]
+                        #[expect(clippy::arithmetic_side_effects)]
                         unsafe {
                             _exit(128 + err)
                         };
@@ -412,7 +413,7 @@ impl WordExp<'_> {
                     if let Err(ref error) = pipe.write_all(b" ") {
                         let err = err2no(error) as i32;
                         // SAFETY: In libc we trust.
-                        #[allow(clippy::arithmetic_side_effects)]
+                        #[expect(clippy::arithmetic_side_effects)]
                         unsafe {
                             _exit(128 + err)
                         };
@@ -557,8 +558,8 @@ impl WordExp<'_> {
     ///
     /// Panics on all errors except Landlock and namespaces which are
     /// optional as they may not be available.
-    #[allow(clippy::cognitive_complexity)]
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::cognitive_complexity)]
+    #[expect(clippy::disallowed_methods)]
     pub fn confine() {
         // SAFETY: Determine user HOME directory.
         // This will be confined by Landlock.
@@ -677,8 +678,8 @@ impl WordExp<'_> {
 
         // SAFETY: Landlock: confine filesystem as read-only.
         // continue on errors as Landlock may not be supported.
-        let mut path_ro = HashSet::default();
-        let mut path_rw = HashSet::default();
+        let mut path_ro = SydHashSet::default();
+        let mut path_rw = SydHashSet::default();
         for ro in [
             "/bin",
             "/dev",
@@ -807,11 +808,15 @@ impl WordExp<'_> {
             }
         }
 
-        const PR_SET_MDWE: nix::libc::c_int = 65;
-        const PR_MDWE_REFUSE_EXEC_GAIN: nix::libc::c_ulong = 1;
-        // SAFETY: set Memory-Deny-Write-Execute attribute.
+        #[cfg(not(any(
+            target_arch = "mips",
+            target_arch = "mips32r6",
+            target_arch = "mips64",
+            target_arch = "mips64r6"
+        )))]
+        // Set Memory-Deny-Write-Execute attribute.
         // continue on errors as MDWE may not be supported.
-        match Errno::result(unsafe { prctl(PR_SET_MDWE, PR_MDWE_REFUSE_EXEC_GAIN, 0, 0, 0) }) {
+        match confine_mdwe(false) {
             Ok(_) => {
                 debug!("ctx": "expand",
                     "msg": "set Memory-Deny-Write-Execute attribute to deny W^X memory");
@@ -835,7 +840,7 @@ impl WordExp<'_> {
         Self::confine_seccomp();
     }
 
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::disallowed_methods)]
     fn confine_seccomp() {
         let mut filter = ScmpFilterContext::new(ScmpAction::Errno(ENOSYS)).expect("create filter");
 
@@ -864,8 +869,8 @@ impl WordExp<'_> {
             "msg": "loaded seccomp filter");
     }
 
-    #[allow(clippy::cognitive_complexity)]
-    #[allow(clippy::disallowed_methods)]
+    #[expect(clippy::cognitive_complexity)]
+    #[expect(clippy::disallowed_methods)]
     fn mount_proc(flags: MsFlags) {
         let mut stack = [0u8; MINI_STACK_SIZE];
         safe_clone(
